#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>

#define _PIOPC_DRV
#include "piopc.h"

#define MAX_PIS	8

#define PIO_CFG_REG	0
#define PIO_AD_REG	2
#define PIO_BD_REG	4
#define PIO_DELAY_REG	6

#define dprintf(...) printk("piopc debug: " __VA_ARGS__)
#define wprintf(...) printk(KERN_WARNING "piopc:" __VA_ARGS__)

typedef struct {
	int used : 1;
	int busy : 1;
	int reserved : 30;

	io_req_t io;
	config_req_t conf;

	struct device *dev;
	struct pcmcia_device *link;

	union {
		struct {
			char led	: 1;
			char pioa_inout	: 1;
			char piob_inout	: 4;
			char reserved	: 2;
		} __attribute__((__packed__)) s;
		unsigned char byte;
	} cfg;
} pio_info;

static const char pio_dev_name[] = "pio";

static pio_info pis[MAX_PIS];
static int Major;
static struct class *pio_class;


/***********
 * Housework functions
 *************************/
static pio_info * pio_info_alloc(void) {
	int i;
	for (i = 0; i < MAX_PIS; ++i) {
		pio_info *info = &pis[i];
		if (!info->used) {
			info->used = 1;
			info->busy = 0;
			return info;
		}
	}
	return NULL;
}

static inline void pio_info_free(pio_info *info) {
	info->used = 0;
}


/****************
 * HW functions
 ******************/
static inline void pio_hw_update_cfg(pio_info *info) {
	outb(info->cfg.byte, info->io.BasePort1 + PIO_CFG_REG);
}

static void pio_hw_set_led(pio_info *info, int light) {
	info->cfg.s.led = !!light;
	pio_hw_update_cfg(info);
}

static void pio_hw_init(pio_info *info) {
	info->cfg.s.led = 0;
	info->cfg.s.pioa_inout = 1;
	info->cfg.s.piob_inout = 1;
	pio_hw_update_cfg(info);
}

static inline unsigned char pio_hw_read(pio_info *info, int regid) {
	return inb(info->io.BasePort1 + regid);
}

static inline void pio_hw_write(pio_info *info, int regid, unsigned char data) {
	outb(data, info->io.BasePort1 + regid);
}

/*************
 * DEV functions
 ******************/
static int pio_dev_open(struct inode *inode, struct file *file) {
	unsigned int id = iminor(inode);
	pio_info *info = &pis[id];

	if (id >= MAX_PIS || !info->used)
		return -ENODEV;
	if (info->busy) return -EBUSY;

	info->busy = 1;
	pio_hw_set_led(info, 1);

	file->private_data = info;

	return 0;
}

static int pio_dev_release(struct inode *inode, struct file *file) {
	pio_info *info = (pio_info*)file->private_data;

	info->busy = 0;
	pio_hw_set_led(info, 0);

	return 0;
}

static ssize_t pio_dev_read(struct file *file, char __user *buf, size_t count, loff_t *pos) {
	pio_info *info = (pio_info*)file->private_data;
	unsigned char piod[2];

	if (count < 2)
		return -EINVAL;

	piod[0] = pio_hw_read(info, PIO_AD_REG);
	piod[1] = pio_hw_read(info, PIO_BD_REG);

	if (copy_to_user(buf, piod, 2))
		return -EFAULT;

	return 2;
}

static ssize_t pio_dev_write(struct file *file, const char __user *buf, size_t count, loff_t *pos) {
	pio_info *info = (pio_info*)file->private_data;
	unsigned char piod[2];

	if (count < 2)
		return -EINVAL;

	if (copy_from_user(piod, buf, 2))
		return -EFAULT;

	pio_hw_write(info, PIO_AD_REG, piod[0]);
	pio_hw_write(info, PIO_BD_REG, piod[1]);

	return 2;
}

static int pio_dev_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) {
	pio_info *info = (pio_info*)file->private_data;
	modconf_t mod;
	int e, mask;

	switch (cmd) {
		case PIOPC_SET_PORTA_MASK:
			if (arg != 0x00 && arg != 0xFF)
				return -EINVAL;
			info->cfg.s.pioa_inout = arg & 1;
			pio_hw_update_cfg(info);
			break;

		case PIOPC_SET_PORTB_MASK:
			if (arg & ~0x0F)
				return -EINVAL;
			info->cfg.s.piob_inout = arg & 0x0F;
			pio_hw_update_cfg(info);
			break;

		case PIOPC_SET_DELAY:
			if (arg > 0xFF)
				return -EINVAL;
			pio_hw_write(info, PIO_DELAY_REG, arg & 0xFF);
			break;

		// TODO: geteri
		case PIOPC_GET_PORTA_MASK:
		case PIOPC_GET_PORTB_MASK:
			if (cmd == PIOPC_GET_PORTA_MASK)
				mask = info->cfg.s.pioa_inout ? 0xFF : 0x00;
			else
				mask = info->cfg.s.piob_inout;

			if (put_user(mask, (int*)arg))
				return -EFAULT;
			break;

		case PIOPC_GET_CFG:
			if (put_user(pio_hw_read(info, PIO_CFG_REG), (unsigned char*)arg))
				return -EFAULT;
			break;

		case PIOPC_SET_VPP:
			mod.Attributes = CONF_VPP1_CHANGE_VALID | CONF_VPP2_CHANGE_VALID;
			mod.Vpp1 = arg;
			mod.Vpp2 = arg;
			if ((e = pcmcia_modify_configuration(info->link, &mod)))
				return e;
			break;

		default:
			return -EINVAL;
	}

	return 0;
}

/*************
 * PCMCIA functions
 ************************/
static int pio_probe(struct pcmcia_device *link) {
	pio_info *info;
	int pio_id;
	int e;

	dprintf("Detected PIO PCMCIA!\n");

	info = pio_info_alloc();
	if (!info) return -ENOMEM;

	pio_id = info - pis;
	link->priv = info;
	info->link = link;

	// card init
	info->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
	info->io.IOAddrLines = 5;
	info->io.BasePort1 = 0;
	info->io.NumPorts1 = 32;
	info->io.NumPorts2 = 0;

	info->conf.Attributes = 0;
	info->conf.IntType = INT_MEMORY_AND_IO;
	info->conf.Vpp = 50;
	info->conf.Present = 0;

	if ((e = pcmcia_request_io(link, &info->io))) {
		dprintf("pcmcia_request_io failed with %d\n", e);
		goto err0;
	}

	if ((e = pcmcia_request_configuration(link, &info->conf))) {
		dprintf("pcmcia_request_configuration failed with %d\n", e);
		goto err0;
	}

	// hw init
	pio_hw_init(info);

	// device file init
	info->dev = device_create(pio_class, &link->dev, MKDEV(Major, pio_id), NULL, "pio%d", pio_id);
	if (IS_ERR(info->dev)) {
		wprintf("Unable to create device for PIO %d; errno = %ld\n", pio_id, PTR_ERR(info->dev));
		info->dev = NULL;
	}

	return 0;

err0:
	pcmcia_disable_device(link);
	pio_info_free(info);
	return e;
}

static void pio_remove(struct pcmcia_device *link) {
	pio_info *info = (pio_info*)link->priv;
	int id = info - pis;

	if (info->dev) device_destroy(pio_class, MKDEV(Major, id));

	pcmcia_disable_device(link);

	pio_info_free(info);
}

/*****************
 * Init c0de
 ********************/
static struct pcmcia_device_id pio_ids[] = {
	PCMCIA_DEVICE_MANF_CARD(0xDC00, 0x00AA),
	PCMCIA_DEVICE_NULL
};
MODULE_DEVICE_TABLE(pcmcia, pio_ids);

static struct file_operations pio_ops = {
	.owner = THIS_MODULE,
	.read = pio_dev_read,
	.write = pio_dev_write,
	.open = pio_dev_open,
	.release = pio_dev_release,
	.ioctl = pio_dev_ioctl
};

static struct pcmcia_driver pio_cs_driver = {
	.owner = THIS_MODULE,
	.drv = {
		.name = "pio_pc"
	},
	.probe = pio_probe,
	.remove = pio_remove,
	.id_table = pio_ids,
	//.suspend, .resume
};

static int __init piopc_init(void) {
	int e;

	pio_class = class_create(THIS_MODULE, "pio");
	if (IS_ERR(pio_class)) {
		e = PTR_ERR(pio_class);
		dprintf("class_create failed with %d\n", e);
		return e;
	}

	Major = e = register_chrdev(0, pio_dev_name, &pio_ops);
	if (e < 0) {
		dprintf("register_chrdev failed with %d\n", e);
		return e;
	}

	if ((e = pcmcia_register_driver(&pio_cs_driver))) {
		dprintf("pcmcia_register_driver failed with %d\n", e);
		return e;
	}

	dprintf("loaded\n");

	return 0;
}

static void __exit piopc_exit(void) {
	pcmcia_unregister_driver(&pio_cs_driver);
	unregister_chrdev(Major, pio_dev_name);
	class_destroy(pio_class);
}

module_init(piopc_init);
module_exit(piopc_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Davor Cihlar");
